/* 
 *	Copyright (C) Mathias Born - May 2001
 *
 *  This file is part of DVD2AVI, a free MPEG-2 decoder
 *	
 *  DVD2AVI is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVI is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "avisynth.h"
#include "huffyuv_a.h"


class MPEG2Source: public IClip {
protected:
  VideoInfo vi;
  CMPEG2Decoder m_decoder;
public:
  MPEG2Source(const char *path, IScriptEnvironment* env);
  MPEG2Source(const char *path);
  ~MPEG2Source();
  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
  void __stdcall GetFrame(int n, unsigned char* buf, const int pitch);
  void __stdcall GetAudio(void* buf, int start, int count, IScriptEnvironment* env) {};
  const VideoInfo& __stdcall GetVideoInfo() { return vi; }
  bool __stdcall GetParity(int n);
};

MPEG2Source::MPEG2Source(const char *path, IScriptEnvironment* env)
{
	if (!m_decoder.Open(path, CMPEG2Decoder::YUV422))
		env->ThrowError("MPEG2Source: couldn't open file");

	memset(&vi, 0, sizeof(vi));
	vi.width = m_decoder.Clip_Width;
	vi.height = m_decoder.Clip_Height;
	vi.pixel_type = VideoInfo::YUY2;
	vi.fps_numerator = m_decoder.VF_FrameRate;
	vi.fps_denominator = 1000;
	vi.num_frames = m_decoder.VF_FrameLimit;
	vi.field_based = false;
}

MPEG2Source::MPEG2Source(const char *path)
{
	memset(&vi, 0, sizeof(vi));
	
	if (m_decoder.Open(path, CMPEG2Decoder::YUV422))
	{
		vi.width = m_decoder.Clip_Width;
		vi.height = m_decoder.Clip_Height;
		vi.pixel_type = VideoInfo::YUY2;
		vi.fps_numerator = m_decoder.VF_FrameRate;
		vi.fps_denominator = 1000;
		vi.num_frames = m_decoder.VF_FrameLimit;
		vi.field_based = false;
	}
}

MPEG2Source::~MPEG2Source()
{
	m_decoder.Close();
}

bool __stdcall MPEG2Source::GetParity(int)
{
	return m_decoder.Field_Order == 1;	// todo: check this
}

PVideoFrame __stdcall MPEG2Source::GetFrame(int n, IScriptEnvironment* env)
{
    PVideoFrame frame = env->NewVideoFrame(vi);
    unsigned char* dst = frame->GetWritePtr();
    const int pitch = frame->GetPitch();

	m_decoder.Decode(dst, n, pitch);

	return frame;
}

void __stdcall MPEG2Source::GetFrame(int n, unsigned char* buf, const int pitch)
{
	m_decoder.Decode(buf, n, pitch);
}

AVSValue __cdecl Create_MPEG2Source(AVSValue args, void*, IScriptEnvironment* env)
{
  const char* path = args[0].AsString();
  return new MPEG2Source(args[0].AsString(), env);
}

MPEG2Source* vf = NULL;
unsigned char* buffer = NULL;
unsigned char* picture = NULL;
VideoInfo pvi;
int pitch = 0;

extern "C" __declspec(dllexport) VideoInfo* __stdcall openMPEG2Source(char* file)
{
	if (vf != NULL) delete vf;
	vf = new MPEG2Source(file);

	pvi = vf->GetVideoInfo();
//	if (picture != NULL) free(picture);
	if (picture != NULL) aligned_free(picture);
//	if (buffer != NULL) free(buffer);
	if (buffer != NULL) aligned_free(buffer);
//	buffer = (unsigned char*)malloc(pvi.height * pvi.width * 2);
	buffer = (unsigned char*)aligned_malloc(pvi.height * pvi.width * 2, 128);

	pitch = (((pvi.width * 2) + 7) / 8) * 8;
	return &pvi;
}

extern "C" __declspec(dllexport) unsigned char* __stdcall getFrame(int frame)
{
	vf->GetFrame(frame, buffer, pitch);
	return buffer;
}

extern "C" __declspec(dllexport) unsigned char* __stdcall getRGBFrame(int frame)
{
	vf->GetFrame(frame, buffer, pitch);
//	if (picture == NULL) picture = (unsigned char*)malloc(pvi.height * pvi.width * 3);
	if (picture == NULL) picture = (unsigned char*)aligned_malloc(pvi.height * pvi.width * 3, 128);
	mmx_UYVYtoRGB24(buffer, picture, buffer + pvi.height * pvi.width * 2 - 1, pvi.width * 2);
	return picture;
}

extern "C" __declspec(dllexport) void __stdcall closeVideo(void)
{
	if (buffer != NULL)
	{
//		free(buffer);
		aligned_free(buffer);
		buffer = NULL;
	}
	if (picture != NULL)
	{
//		free(picture);
		aligned_free(picture);
		picture = NULL;
	}
	if (vf != NULL)
	{
		delete vf;
		vf = NULL;
	}
}