/* 
 *	Copyright (C) Chia-chen Kuo - April 2001
 *
 *  This file is part of DVD2AVI, a free MPEG-2 decoder
 *	
 *  DVD2AVI is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVI is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "vfapi.h"

#undef MPEG2DEC_API
#define MPEG2DEC_API __stdcall


extern "C" {
  HRESULT MPEG2DEC_API vfGetPluginInfo(LPVF_PluginInfo info);
  HRESULT MPEG2DEC_API vfGetPluginFunc(LPVF_PluginFunc func);

  HRESULT MPEG2DEC_API open_file(char *path, LPVF_FileHandle out);
  HRESULT MPEG2DEC_API close_file(VF_FileHandle p);
  HRESULT MPEG2DEC_API get_file_info(VF_FileHandle in, LPVF_FileInfo out);
  HRESULT MPEG2DEC_API get_stream_info(VF_FileHandle in, DWORD s, void *out);
  HRESULT MPEG2DEC_API read_data(VF_FileHandle in, DWORD s, void *out);
}

HRESULT MPEG2DEC_API vfGetPluginInfo(LPVF_PluginInfo info)
{
	if (info == NULL)
		return VF_ERROR;

	if (info->dwSize != sizeof(VF_PluginInfo))
		return VF_ERROR;

	info->dwAPIVersion = 1;
	info->dwVersion = 1;
	info->dwSupportStreamType = VF_STREAM_VIDEO;

	strcpy(info->cPluginInfo, "DVD2AVI Project File Reader 1.76");
	strcpy(info->cFileType, "DVD2AVI Project File (*.d2v)|*.d2v");

	return VF_OK;
}

HRESULT MPEG2DEC_API vfGetPluginFunc(LPVF_PluginFunc func)
{
	if (func == NULL)
		return VF_ERROR;

	if (func->dwSize != sizeof(VF_PluginFunc))
		return VF_ERROR;

	func->OpenFile = open_file;
	func->CloseFile = close_file;
	func->GetFileInfo = get_file_info;
	func->GetStreamInfo = get_stream_info;
	func->ReadData = read_data;

	return VF_OK;
}

HRESULT MPEG2DEC_API open_file(char *path, LPVF_FileHandle out)
{
	CMPEG2Decoder **d2v = (CMPEG2Decoder**)out;
	*d2v = new CMPEG2Decoder();
	CMPEG2Decoder *dec = *d2v;

	if (dec == NULL) return VF_ERROR;

	if (!dec->Open(path, CMPEG2Decoder::RGB24))
	{
		delete dec;
		*d2v = NULL;
		return VF_ERROR;
	}

	return VF_OK;
}

HRESULT MPEG2DEC_API close_file(VF_FileHandle in)
{
	CMPEG2Decoder *dec = (CMPEG2Decoder*)in;

	dec->Close();
	delete dec;

	return VF_OK;
}

HRESULT MPEG2DEC_API get_file_info(VF_FileHandle in, LPVF_FileInfo out)
{
	if (out == NULL)
		return VF_ERROR;

	if (out->dwSize != sizeof(VF_FileInfo))
		return VF_ERROR;

	out->dwHasStreams = VF_STREAM_VIDEO;

	return VF_OK;
}

HRESULT MPEG2DEC_API get_stream_info(VF_FileHandle in, DWORD stream, void *out)
{
	LPVF_StreamInfo_Video info;

	if (stream != VF_STREAM_VIDEO)
		return VF_ERROR;

	CMPEG2Decoder *dec = (CMPEG2Decoder*)in;
	info = (LPVF_StreamInfo_Video)out;

	if (info == NULL)
		return VF_ERROR;
	
	if (info->dwSize != sizeof(VF_StreamInfo_Video))
		return VF_ERROR;

	info->dwLengthL = dec->VF_FrameLimit;
	info->dwRate = dec->VF_FrameRate;
	info->dwScale = 1000;
	info->dwWidth = dec->Clip_Width;
	info->dwHeight = dec->Clip_Height;
	info->dwBitCount = 24;

	return VF_OK;
}
	
HRESULT MPEG2DEC_API read_data(VF_FileHandle in, DWORD stream, void *out)
{
	unsigned char *dst;
	LPVF_ReadData_Video data;

	if (stream != VF_STREAM_VIDEO)
		return VF_ERROR;

	CMPEG2Decoder *dec = (CMPEG2Decoder*)in;
	data = (LPVF_ReadData_Video)out;
	dst = (unsigned char *)data->lpData;

	if (data == NULL)
		return VF_ERROR;
	
	if (data->dwSize != sizeof(VF_ReadData_Video))
		return VF_ERROR;

	dec->Decode(dst, data->dwFrameNumberL, data->lPitch);

	return VF_OK;
}
